/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

/* 
 * Note to regenerate the Makefile and configure script:
 * aclocal
 * autoconf
 * touch NEWS README AUTHORS ChangeLog
 * automake -a
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <math.h>
#include <limits.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <wctype.h>
#include <wchar.h>
#include <locale.h>
#include <langinfo.h>

#if defined HAVE_LIBBOOST_REGEX

#define UNICODE
#include <boost/regex.h>

#else 

#include <sys/types.h>
#include <regex.h>

#endif

/* the following defines set up a tradeoff between
   modelling accuracy and memory requirements - season to taste */

/* digram digitization: avg loss of precision = 0.01 * token size */
#define DIGITIZE_DIGRAMS
/* lambda digitization: avg loss of precision = 0.01 */
#define DIGITIZE_LAMBDA
/* learner.hash digitization: avg loss of precision = 0.01 */
#define DIGITIZE_LWEIGHTS
/* tolerance for the error in divergence - 
   more accuracy means slower learning. */
#define TOL 0.05

#include "dbacl.h" /* make sure this is last */

/* global variables */

hash_bit_count_t default_max_hash_bits = 15;
hash_count_t default_max_tokens = (1<<15);

hash_bit_count_t decimation;

Dirichlet dirichlet;

Learner learner;
Category cat[MAX_CAT];
category_count_t cat_count = 0;
int filter[MAX_CAT];
category_count_t filter_count = 0;
Regex re[MAX_RE];
regex_count_t regex_count = 0;
Empirical empirical;

char *textbuf = NULL;
charbuf_len_t textbuf_len = 0;
wchar_t *wc_textbuf = NULL;
charbuf_len_t wc_textbuf_len = 0;
char *aux_textbuf = NULL;
charbuf_len_t aux_textbuf_len = 0;

token_order_t ngram_order = 1;

MBOX_State mbox;
XML_State xml;

/* for option processing */
extern char *optarg;
extern int optind, opterr, optopt;

options_t options = 0;
char *title = "";

/* string hash function */
extern unsigned long int hash( unsigned char *k, 
			       unsigned long int length, 
			       unsigned long int initval);

int exit_code = 0; /* default */
int hashfull_warning = 0;
int overflow_warning = 0;
int digramic_overflow_warning = 0;
int skewed_constraints_warning = 0;
/***********************************************************
 * MISCELLANEOUS FUNCTIONS                                 *
 ***********************************************************/

void usage(char **argv) {
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "dbacl [-vniNR] [-T type] -c CATEGORY [-c CATEGORY]...\n");
  fprintf(stderr, 
	  "      [-f KEEP]... [FILE]...\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      classifies FILE or STDIN using CATEGORY, optionally\n");
  fprintf(stderr, 
	  "      removing all lines which don't fit the category KEEP.\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "dbacl [-vnirND] [-h size] [-T type] -l CATEGORY \n");
  fprintf(stderr, 
	  "      [-g regex]... [FILE]...\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      builds a maximum entropy model from the words in FILE or STDIN\n");
  fprintf(stderr, 
	  "      or concatenated regex submatches if using the -g option.\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "dbacl -V\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      prints program version.\n");
}

digitized_weight_t digitize_a_weight(weight_t w) {
  if( w < -320 ) {
    return DIGITIZED_WEIGHT_MIN;
  } else if( 320 < w ) {
    return DIGITIZED_WEIGHT_MAX;
  } else {
    return 100 * w;
  }
}


/***********************************************************
 * EMPIRICAL DISTRIBUTION OF TOKENS                        *
 ***********************************************************/

/* initialize global learner object */
void init_empirical() {

    /* some constants */
    empirical.max_tokens = default_max_tokens;
    empirical.max_hash_bits = default_max_hash_bits;
    empirical.full_token_count = 0;
    empirical.unique_token_count = 0;
    empirical.track_features = 0;
    empirical.feature_stack_top = 0;

    /* allocate room for hash */
    empirical.hash = calloc(empirical.max_tokens, sizeof(h_item));
    if( !empirical.hash ) {
	fprintf(stderr, 
		"error: not enough memory? I couldn't allocate %li bytes\n", 
		(sizeof(h_item) * ((long int)empirical.max_tokens)));
	exit(0);
    }
}

void clear_empirical() {
    token_stack_t i;

    if( empirical.track_features ) {
	/* this may actually be slower than a global memset */ 
	for(i = 0; i < empirical.feature_stack_top; i++) {
	    memset(empirical.feature_stack[i], 0, sizeof(h_item));
	}
    } else {
	memset(empirical.hash, 0, sizeof(h_item) * empirical.max_tokens);
    }

    empirical.full_token_count = 0;
    empirical.unique_token_count = 0;
    empirical.feature_stack_top = 0;
}

h_item *find_in_empirical(unsigned long int id) {
  register h_item *i, *loop;
  /* start at id */
  i = loop = &empirical.hash[id & (empirical.max_tokens - 1)];

  while( FILLEDP(i) ) {
    if( EQUALP(i->id,id) ) {
      return i; /* found id */
    } else {
      i++; /* not found */
      /* wrap around */
      i = (i >= &empirical.hash[empirical.max_tokens]) ? empirical.hash : i; 
      if( i == loop ) {
	return NULL; /* when hash table is full */
      }
    }
  }

  /* empty slot, so not found */

  return i; 
}

/* calculates the entropy of the empirical measure */
score_t empirical_entropy() {
    hash_count_t i;
    score_t e = 0.0;

    if( empirical.track_features && 
	(empirical.feature_stack_top < MAX_TOKEN_LINE_STACK) ) {
	for(i = 0; i < empirical.feature_stack_top; i++) {
	    e += ((score_t)empirical.feature_stack[i]->count) * 
	      log((score_t)empirical.feature_stack[i]->count);
	}
    } else {
	for(i = 0; i < empirical.max_tokens; i++) {
	    if( FILLEDP(&empirical.hash[i]) ) {
		e += ((score_t)empirical.hash[i].count) * 
		  log((score_t)empirical.hash[i].count);
	    }
	}
    }
    e = e/empirical.full_token_count - log((score_t)empirical.full_token_count);

    return -e;
}

/***********************************************************
 * CATEGORY FUNCTIONS                                      *
 ***********************************************************/

/* initialize to zero. Filename specified elsewhere */
void init_category(Category *cat) {
    char *p;

    cat->model_full_token_count = 0;
    cat->model_unique_token_count = 0;
    cat->score = 0.0;
    cat->complexity = 0;
    cat->max_order = 0;
    if( (p = strrchr(cat->filename, '/')) ) {
	cat->filename = p + 1; /* only keep basename */
    }
    cat->retype = 0;
    cat->model_type = simple;
}

/* turns purely random text into a category of its own */
void init_purely_random_text_category(Category *cat) {
  alphabet_size_t i, j;
  weight_t z = -log((double)ASIZE - 1.0);

#if defined DIGITIZE_DIGRAMS
  digitized_weight_t zz = PACK_DIGRAMS(z);
#endif
    
  for(i = 1; i < ASIZE; i++) {
    for(j = 1; j < ASIZE; j++) {
#if defined DIGITIZE_DIGRAMS
      cat->dig[i][j] = zz;
#else
      cat->dig[i][j] = z;
#endif
    }
  }

  z = -log((double)ASIZE - 2.0);
#if defined DIGITIZE_DIGRAMS
  zz = PACK_DIGRAMS(z);
#endif
  for(j = 1; j < ASIZE; j++) {
#if defined DIGITIZE_DIGRAMS
    cat->dig[DIAMOND][j] = zz;
#else
    cat->dig[DIAMOND][j] = z;
#endif
  } 

  /* not needed: set DIAMOND-DIAMOND score for completeness only */

#if defined DIGITIZE_DIGRAMS
  cat->dig[DIAMOND][DIAMOND] = DIGITIZED_WEIGHT_MIN;
#else
  cat->dig[DIAMOND][DIAMOND] = log(0.0);
#endif

  cat->logZ = 0.0;
  cat->hash = NULL;
  cat->model_type = simple;
  cat->max_order = 1;
  options |= (1<<OPTION_NOREGEX); 
}

c_item *find_in_category(Category *cat, unsigned long int id) {
    register c_item *i, *loop;

    if( cat->hash ) {
	/* start at id */
	i = loop = &cat->hash[id & (cat->max_tokens - 1)];

	while( FILLEDP(i) ) {
	    if( EQUALP(i->id,id) ) {
		return i; /* found id */
	    } else {
		i++; /* not found */
		/* wrap around */
		i = (i >= &cat->hash[cat->max_tokens]) ? cat->hash : i; 
		if( i == loop ) {
		    return NULL; /* when hash table is full */
		}
	    }
	}
	return i;
    } else {
	return NULL;
    }
}

/* for each loaded category, this calculates the score */
void score_word(char *tok, token_order_t r, regex_count_t re) {
  category_count_t i;
  alphabet_size_t pp, pc;
  hash_value_t id;
  char *q;
  register c_item *k;
  h_item *h = NULL;

  if( *(tok + 2) ) { /* DIAMOND-DIAMOND is empty string */

    id = (hash_value_t)hash(tok, strlen(tok), 0);

    if( (options & (1<<OPTION_CALCENTROPY)) && (r == 1) ) {
      /* add the token to the hash */

      h = find_in_empirical(id);
      if( h ) {
 	if( FILLEDP(h) ) {
	  h->count += ( h->count < K_TOKEN_COUNT_MAX ) ? 1 : 0;
 	} else {
 	  if( /* !FILLEDP(i) && */
 	     ((100 * empirical.unique_token_count) <
 	      (HASH_FULL * empirical.max_tokens) )) {
 	    /* fill the empirical hash */
 	    SET(h->id,id);
	    empirical.unique_token_count += 
	      ( empirical.unique_token_count < K_TOKEN_COUNT_MAX ) ? 1 : 0;
	    h->count += ( h->count < K_TOKEN_COUNT_MAX ) ? 1 : 0;
 	  } else {
 	    /* hash full */
 	    h = NULL; 
 	    if( !hashfull_warning ) {
 	      fprintf(stderr,
 		      "warning: empirical hash full, calculation may be skewed. " 
 		      "Try option -h %d\n", 
 		      (empirical.max_hash_bits + 1));
 	      hashfull_warning = 1;
 	    }
 	    return; /* pretend word doesn't exist */
 	  }

	  if( empirical.track_features &&
	      (empirical.feature_stack_top < MAX_TOKEN_LINE_STACK)) {
	    empirical.feature_stack[empirical.feature_stack_top++] = h;
	  }
 	}
	empirical.full_token_count += 
	  ( empirical.full_token_count < K_TOKEN_COUNT_MAX ) ? 1 : 0;
	
      }
    }

    /* now do scoring for all available categories */
    for(i = 0; i < cat_count; i++) {

      switch(cat[i].model_type) {
      case simple:
	/* see if this is for us */
	if( ((re == 0) && (r == 1)) ||
	    ((re > 0) && (cat[i].retype & (1<<(re-1)))) ) {
	  /* if token found, add its lambda weight */
	  k = find_in_category(&cat[i], id);
	  if( k ) {
	    cat[i].score += UNPACK_LAMBDA(k->lam);
	  }

	  /* now compute the reference weight from digram model */
	  pp = (unsigned char)*tok;
	  q = tok + 1;
	  while( *q ) {
	    if( *q == '\r' ) {
	      q++;
	      continue;
	    }
	    pc = (unsigned char)*q;
	    cat[i].score += UNPACK_DIGRAMS(cat[i].dig[pp][pc]);
	    pp = pc;
	    q++;
	  }

	  /* don't forget the normalizing constant */
	  cat[i].score -= cat[i].logZ;      

	  /* update complexity for this category */
	  cat[i].complexity++;

	  /* add correction factor for multinomial case */
	  if(h) {
	    cat[i].score += log((weight_t)cat[i].complexity) - 
	      log((weight_t)h->count);
	  }
	}
	break;

      case sequential:
      default:

	/* see if this is for us */
	if( ((re == 0) && !cat[i].retype) ||
	    ((re > 0) && (cat[i].retype & (1<<(re-1)))) ) {

	  /* if token found, add its lambda weight */
	  k = find_in_category(&cat[i], id);
	  if( k ) {
	    cat[i].score += UNPACK_LAMBDA(k->lam);
	  }

	  if( r == 1 ) {
	    /* now compute the reference weight from digram model */
	    pp = (unsigned char)*tok;
	    q = tok + 1;
	    while( *q ) {
	      if( *q == '\r' ) {
		q++;
		continue;
	      }
	      pc = (unsigned char)*q;
	      cat[i].score += UNPACK_DIGRAMS(cat[i].dig[pp][pc]);
	      pp = pc;
	      q++;
	    }
	  }

	  if( r == cat[i].max_order ) {
	    /* don't forget the normalizing constant */
	    cat[i].score -= cat[i].logZ;      

	    /* update complexity for this category */
	    cat[i].complexity++;
	  }
	}
	break;

      }

    }
  }
}

void line_score_single_category() {
    score_t score = 0.0;

    if( *textbuf ) {
      if( options & (1<<OPTION_POSTERIOR) ) {
	//score = empirical_entropy();
	fprintf(stdout, "%s", textbuf);
      }
      if( options & (1<<OPTION_SCORES) ) {
	if( options & (1<<OPTION_VERBOSE) ) {
	  score = -(cat[0].score/cat[0].complexity)/log(2.0);
	  fprintf(stdout, "%s %6.2" FMT_printf_score_t " * %-4d %s", 
		  cat[0].filename, score, 
		  cat[0].complexity, textbuf);
	} else {
	  fprintf(stdout, "%s %7.2" FMT_printf_score_t " %s", 
		  cat[0].filename, -cat[0].score, textbuf); 
	}
      }

    }
    cat[0].score = 0.0;
    cat[0].complexity = 0;

    if( options & (1<<OPTION_CALCENTROPY) ) {
      clear_empirical();
    }
}

void line_score_multiple_categories() {
  category_count_t i;
  int map;
  score_t c, cmax;

  /* find MAP */
  cmax = cat[0].score; 
  map = 0;
  for(i = 1; i < cat_count; i++) {
    if(cmax < cat[i].score) {
      cmax = cat[i].score;
      map = i;
    }
  }

  if(options & (1<<OPTION_POSTERIOR) ) {
    /* compute probabilities given exclusive choices */
    c = 0.0;
    for(i = 0; i < cat_count; i++) {
      c += exp((cat[i].score - cmax));
    }
      
    if( *textbuf ) {
      for(i = 0; i < cat_count; i++) {
	fprintf(stdout, "%s %6.2" FMT_printf_score_t "%% ", 
		cat[i].filename, 100 * exp((cat[i].score - cmax))/c);
      }
      fprintf(stdout, "%s", textbuf);
    }
  } else if( options & (1<<OPTION_SCORES) ) {
    /* display normalized divergence scores */
    if( *textbuf ) {
      for(i = 0; i < cat_count; i++) {
	if( options & (1<<OPTION_VERBOSE) ) {
	  fprintf(stdout, "%s %6.2" FMT_printf_score_t " * %-4d ", 
		  cat[i].filename, 
		  -cat[i].score/cat[i].complexity, 
		  cat[i].complexity);
	} else {
	  fprintf(stdout, "%s %6.2" FMT_printf_score_t " ", 
		  cat[i].filename, -cat[i].score);
	}
      }
      fprintf(stdout, "%s", textbuf);
    }
  } else {
    /* prune the text which doesn't fit */
    for(i = 0; i < filter_count; i++) {
      if( cat[map].score <= cat[filter[i]].score ) {
	fprintf(stdout, "%s", textbuf);
	break;
      }
    }
  }
    
  /* clean up for next line */
  for(i = 0; i < cat_count; i++) {
    cat[i].score = 0.0;
    cat[i].complexity = 0;
  }
  if( options & (1<<OPTION_CALCENTROPY) ) {
    clear_empirical();
  }
}

void score_single_category() {
  //score_t score = 0.0;

  if( options & (1<<OPTION_POSTERIOR) ) {
      //score = empirical_entropy();
    //fprintf(stdout, "n/a\n");
  }

  if( options & (1<<OPTION_APPEND) ) {
    fprintf(stdout, "\nscores ");
  }

  if( options & (1<<OPTION_SCORES) ) {
    if( options & (1<<OPTION_VERBOSE) ) {
      fprintf(stdout, "cross_entropy %.2" FMT_printf_score_t \
                      " bits complexity %d\n", 
	      -(cat[0].score/cat[0].complexity) / log(2.0), cat[0].complexity);
    } else {
      fprintf(stdout, "score %.2" FMT_printf_score_t "\n", -cat[0].score);
    }
  }
  exit_code = 1;
}

void score_multiple_categories() {
  category_count_t i;
  score_t c, cmax;

  if( options & (1<<OPTION_APPEND) ) {
    fprintf(stdout, "\nscores ");
  }

  /* find MAP */
  cmax = cat[0].score; 
  exit_code = 0;
  for(i = 1; i < cat_count; i++) {
    if(cmax < cat[i].score) {
      cmax = cat[i].score;
      exit_code = i;
    }
  }

  if( options & (1<<OPTION_POSTERIOR) ) {
    /* here we compute probabilities given exclusive choices */
    c = 0.0;
    for(i = 0; i < cat_count; i++) {
      c += exp((cat[i].score - cmax));
    }

    for(i = 0; i < cat_count; i++) {
      fprintf(stdout, "%s %5.2" FMT_printf_score_t "%% ", 
	      cat[i].filename, 100 * exp((cat[i].score - cmax))/c);
    }
    fprintf(stdout, "\n");
  } else if( options & (1<<OPTION_SCORES) ) {
    /* display logarithmic score */
    for(i = 0; i < cat_count; i++) {
      if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stdout, "%s %5.2" FMT_printf_score_t " * %-d ", 
		cat[i].filename, (-cat[i].score/cat[i].complexity)/log(2.0), 
		cat[i].complexity);
      } else {
	fprintf(stdout, "%s %5.2" FMT_printf_score_t " ", 
		cat[i].filename, -cat[i].score/log(2.0));
      }
    }
    fprintf(stdout, "\n");
  } else if( options & (1<<OPTION_VERBOSE) ) {
    fprintf(stdout, "%s\n", cat[exit_code].filename); 
  }

  exit_code++; /* make number between 1 and cat_count+1 */
}

/***********************************************************
 * FILE MANAGEMENT FUNCTIONS                               *
 ***********************************************************/

/* used by both load and save functions */
#define MAGIC_BUFSIZE 256
#define MAGIC1    "# dbacl version "VERSION" category %s %s\n"
#define MAGIC2_i  "# model entropy: %" FMT_scanf_score_t \
                  " logZ: %" FMT_scanf_score_t " max_order: %" \
                  FMT_token_order_t " type: %s\n"
#define MAGIC2_o  "# model entropy: %" FMT_printf_score_t \
                  " logZ: %" FMT_printf_score_t " max_order: %" \
                  FMT_token_order_t " type: %s\n"
#define MAGIC3    "# hash_size: %" FMT_hash_bit_count_t \
                  " full_token_count: %d unique_token_count: %d\n"
#define MAGIC4    "# case sensitive\n"
#define MAGIC5_i  "# regex %s\n"
#define MAGIC5_o  "# regex %s||%s\n"
#define RESTARTPOS 8
#define MAGIC6    "#\n"
#define MAGIC7    "# wide characters\n"

/* loads a category hash */
error_code_t load_category(Category *cat) {
  hash_count_t i, j;
  regex_count_t c;
  char buf[MAGIC_BUFSIZE];
  char scratchbuf[MAGIC_BUFSIZE];
  char *p;

  FILE *input;

#if defined HAVE_LIBBOOST_REGEX
  charbuf_len_t r;
#endif

  if( (input = fopen(cat->filename, "rb")) ) {
    fgets(buf, MAGIC_BUFSIZE, input);
    if( strncmp(buf, MAGIC1, 16 + strlen(VERSION)) ) {
      fprintf(stderr, 
	      "error: not a dbacl version "VERSION" category file [%s]\n", 
	      cat->filename);
      return 0;
    } 

    init_category(cat); /* changes filename */

    fgets(buf, MAGIC_BUFSIZE, input);
    sscanf(buf, MAGIC2_i, &cat->divergence, &cat->logZ, 
	   &cat->max_order, scratchbuf);
    if( scratchbuf[0] == 'm' ) {
      cat->model_type = simple;
    } else {
      cat->model_type = sequential;
    }
    fgets(buf, MAGIC_BUFSIZE, input);
    sscanf(buf, MAGIC3, &cat->max_hash_bits, 
	   &cat->model_full_token_count, &cat->model_unique_token_count);
    cat->max_tokens = (1<<cat->max_hash_bits);

    /* see if there are any regexes */
    fgets(buf, MAGIC_BUFSIZE, input);
    while(1) {
      if( strncmp(buf, MAGIC6, 2) == 0 ) {
	break;
      } else if( strncmp(buf, MAGIC5_i, 8) == 0 ) {

	/* set up the submatch bitmap */
	re[regex_count].submatches |= 0;

	if( (p = strrchr(buf + RESTARTPOS, '|')) && ( *(--p) == '|') ) {
	  /* assume string ends in ||12345, use as bitmap */
	  *p = '\0';
	  for(p += 2; *p; p++) {
	    /* assume ascii number positions */
	    if( (*p > '9') || (*p < '1')) {
	      if( *p != '\n' ) {
		fprintf(stderr, 
			"warning: could not decode || suffix for %s while loading %s\n", 
			buf + RESTARTPOS, cat->filename);
	      }
	    } else {
	      re[regex_count].submatches |= (1<<(*p - '0'));
	    }
	  }
	} else { /* no bitmap specified */
	  re[regex_count].submatches = ~0;
	}

#if defined HAVE_LIBBOOST_REGEX
	/* boost regexes accept wide characters */
	    
	r = strlen(buf + RESTARTPOS);
	re[regex_count].string = malloc((r + 1) * sizeof(wchar_t));
	/* note: skip the trailing newline */
	r = mbstowcs(re[regex_count].string, buf + RESTARTPOS, r - 1); 
	if( r <= 0 ) {
	  fprintf(stderr, 
		  "error: couldn't convert regular expression '%s'.\n",
		  buf + RESTARTPOS);
	  exit(0); /* no point in going on */
	} else {
	  re[regex_count].string[r] = L'\0';
	}
	/* now see if we already have this string */
	for(c = 0; c < regex_count; c++) {
	  if( wcscmp(re[c].string, re[regex_count].string) == 0 ) {
	    break;
	  }
	}
	if( c < regex_count ) {
	  /* we found this regex, no need to repeat it */
	  free(re[regex_count].string);
	} else {
	  /* compile the regex (uses regcompW) */
	  if( regcomp(&re[regex_count].regex,
		      re[regex_count].string, REG_EXTENDED) != 0 ) {
	    fprintf(stderr, 
		    "error: could not compile regular expression '%ls' for %s.\n", 
		    re[regex_count].string, cat->filename);
	    exit(0);
	  } else {
	    regex_count++;
	    if( regex_count >= MAX_RE ) { 
	      fprintf(stderr, "error: too many regular expressions\n");
	      exit(0); /* no point in going on */
	    }
	  }
	}
#else
	/* remove trailing newline */
	if( buf[strlen(buf) - 1] == '\n' ) { 
	  buf[strlen(buf) - 1] = '\0'; 
	}
	/* GNU regexes use ordinary strings */
	/* it's a regex - see if we've got it already */
	for(c = 0; c < regex_count; c++) {
	  if( strcmp(re[c].string, buf + RESTARTPOS) == 0 ) {
	    break;
	  }
	}
	if( c >= regex_count ) { /* not found */
	  /* add it to our list */
	  re[regex_count].string = strdup(buf + RESTARTPOS);
	  /* and compile the regex */
	  if( regcomp(&re[regex_count].regex, 
		      re[regex_count].string, REG_EXTENDED) != 0 ) {
	    fprintf(stderr, 
		    "error: could not compile regular expression '%s' for %s.\n", 
		    re[regex_count].string, cat->filename);
	    exit(0); /* not much point going on */
	  } else {
	    regex_count++;
	    if( regex_count >= MAX_RE ) { 
	      fprintf(stderr, "error: too many regular expressions\n");
	      exit(0); /* no point in going on */
	    }
	  }
	}
#endif
	/* now flag the regex as taken */
	cat->retype |= (1<<c);

      } else if( strncmp(buf, MAGIC4, 10) == 0) {
	/* it's the case sensitive flag */
	options |= (1<<OPTION_CASEN);
      } else if( strncmp(buf, MAGIC7, 5) == 0) {
	options |= (1<<OPTION_I18N);
      }

      /* finished with current line, get next one */
      fgets(buf, MAGIC_BUFSIZE, input); 
    }

    /* if this category did not register a regex, it wants
       the default processing, so we flag this */
    if( !cat->retype ) {
      options |= (1<<OPTION_NOREGEX);
    }

    /* read character frequencies */
    fread(cat->dig, SIZEOF_DIGRAMS, ASIZE * ASIZE, input);

    /* allocate hash table */
    if( !(cat->hash = malloc(sizeof(c_item) * cat->max_tokens)) ) {
      fprintf(stderr, "error: not enough memory for category %s\n", 
	      cat->filename);
      return 0;
    }
    /* read in hash table */
    i = cat->max_tokens;
    j = 0;
    while( j < i ) {
      j += fread(cat->hash + j, sizeof(c_item), i, input);
    }
    fclose(input);
  } else {
    fprintf(stderr, "error: cannot open file for reading %s\n", 
	    cat->filename);
    return 0;
  }

  return 1;
}

/* writes the learner to a file for easily readable category */
error_code_t save_learner() {
  alphabet_size_t i, j;
  hash_count_t t;
  regex_count_t c;
  FILE *output;
  c_item ci;
  char buf[MAGIC_BUFSIZE];

  char smb[MAX_SUBMATCH+1];
  token_order_t s;
  char *p;

#if defined DIGITIZE_DIGRAMS
  short int shval;
#else
  weight_t shval;
#endif

  if( options & (1<<OPTION_VERBOSE) ) {
    fprintf(stdout, "saving category to file %s\n", learner.filename);
  }
  
  /* don't overwrite data files */
  if( (output = fopen(learner.filename, "r")) ) {
    /* output file exists already */
    fgets(buf, MAGIC_BUFSIZE, output);
    if( strncmp(buf, MAGIC1, 10) != 0 ) {
      fprintf(stderr,
	      "error: the file %s is already used for something, "
	      "use another filename. Nothing written.\n", learner.filename);
      fclose(output);
      exit(0);
    } else {
      /* it's an existing category file */
      fclose(output);
    }
  }
  
  if( (output = fopen(learner.filename, "wb")) ) {
    /* print out standard headers */
    fprintf(output, MAGIC1, learner.filename, 
	    (options & (1<<OPTION_REFMODEL)) ? "(ref)" : "");
    fprintf(output, 
	    MAGIC2_o, learner.divergence, learner.logZ, learner.max_order,
	    (options & (1<<OPTION_MULTINOMIAL)) ? "multinomial" : "hierarchical" );
    fprintf(output, MAGIC3, learner.max_hash_bits, 
	    learner.full_token_count, learner.unique_token_count);

    /* print out any regexes we might need */
    for(c = 0; c < regex_count; c++) {
      /* write the bitmap */
      for(p = smb, s = 1; s <= MAX_SUBMATCH; s++) {
	if( re[c].submatches & (1<<s) ) {
	  *p++ = s + '0';
	}
      }
      *p = '\0';

#if defined HAVE_LIBBOOST_REGEX
      /* does this work? */
      fprintf(output, MAGIC5_o, re[c].string, smb);
#else
      fprintf(output, MAGIC5_o, re[c].string, smb);
#endif
    }
    /* this is optional too */
    if( options & (1<<OPTION_CASEN) ) {
      fprintf(output, MAGIC4);
    }
    if( options & (1<<OPTION_I18N) ) {
      fprintf(output, MAGIC7);
    }
    fprintf(output, MAGIC6); 

    /* end of readable stuff */

    /* character frequencies */
    for(i = 0; i < ASIZE; i++) {
      for(j = 0; j < ASIZE; j++) {
	shval = PACK_DIGRAMS(learner.dig[i][j]);
	while(fwrite(&shval, SIZEOF_DIGRAMS, 1, output) < 1);
      }
    }

    /* token/feature weights */
    for(t = 0; t < learner.max_tokens; t++) {
      /* write each element so that it's easy to read back in a c_item array */
      SET(ci.id,learner.hash[t].id);
#if defined DIGITIZE_LAMBDA
      ci.lam = (digitized_weight_t)((learner.hash[t].lam) * 100);
      if( fabs((learner.hash[t].lam)) > 320 ) {
	fprintf(stderr,
		"warning: digitized an extreme lambda value - results may be unusable\n");
      } 
#else
      ci.lam = (learner.hash[t].lam);
#endif
      while(fwrite(&ci, sizeof(ci), 1, output) < 1); 
    }
    fclose(output);
  } else {
    fprintf(stderr, "error: cannot open file for writing %s\n", 
	    learner.filename);
    return 0;
  }

  return 1;
}

/***********************************************************
 * LEARNER FUNCTIONS                                       *
 ***********************************************************/

l_item *find_in_learner(unsigned long int id) {
    register l_item *i, *loop;
    /* start at id */
    i = loop = &learner.hash[id & (learner.max_tokens - 1)];

    while( FILLEDP(i) ) {
	if( EQUALP(i->id,id) ) {
	    return i; /* found id */
	} else {
	    i++; /* not found */
	    /* wrap around */
	    i = (i >= &learner.hash[learner.max_tokens]) ? learner.hash : i; 
	    if( i == loop ) {
		return NULL; /* when hash table is full */
	    }
	}
    }

    /* empty slot, so not found */

    return i; 
}

/* places the token in the global hash and writes the
   token to a temporary file for later, then updates
   the digram frequencies */
void hash_word_and_learn(char *tok, token_order_t r, regex_count_t re) {
  hash_value_t id;
  l_item *i;
  alphabet_size_t p,q;

  if( *(tok + 2) ) { /* DIAMOND-DIAMOND is empty string - ignore */

    if( overflow_warning ) {
      return; /* for there be troubles ahead */
    }

    if( (options & (1<<OPTION_DECIMATE)) &&
	(rand() > (RAND_MAX>>decimation)) ) {
      return; 
    }

    id = (hash_value_t)hash(tok, strlen(tok), 0);
    i = find_in_learner(id);
    if( i ) {
      if( FILLEDP(i) ) {


      } else if( /* !FILLEDP(i) && */
		((100 * learner.unique_token_count) < 
		 (HASH_FULL * learner.max_tokens) )) { 
	/* fill the hash and write to file */
	SET(i->id,id);

	if( learner.unique_token_count < K_TOKEN_COUNT_MAX )
	  { learner.unique_token_count++; } else { overflow_warning = 1; }

	i->order = (options & (1<<OPTION_MULTINOMIAL)) ? 1 : r;

	/* order accounting */
	learner.max_order = (learner.max_order < i->order) ? 
	  i->order : learner.max_order;

	if( learner.fixed_order_unique_token_count[i->order] < K_TOKEN_COUNT_MAX ) 
	  { learner.fixed_order_unique_token_count[i->order]++; } else 
	    { overflow_warning = 1; }

     	if( (options & (1<<OPTION_DEBUG)) ) {
     	  fprintf(stdout, "match %s %d\n", tok, i->order);
     	}

	/* now save token to temporary file */
	fputs(tok, learner.tmp);
	fputc(-1, learner.tmp);
      }

      if( i->count < K_TOKEN_COUNT_MAX ) 
	{ i->count++; } else { overflow_warning = 1; }

      if( learner.full_token_count < K_TOKEN_COUNT_MAX )
	{ learner.full_token_count++; } else 
	  { /* this number is just cosmetic */ }

      if( learner.fixed_order_token_count[i->order] < K_TOKEN_COUNT_MAX ) 
	{ learner.fixed_order_token_count[i->order]++; } else 
	  { skewed_constraints_warning = 1; }

    }

    if( digramic_overflow_warning ) {
      return;
    }

    /* now update digram frequency counts */
    p = *tok++;
    /* finally update character frequencies */
    while( *tok ) {
      q = (unsigned char)*tok;
      if( learner.dig[p][q] < K_TOKEN_COUNT_MAX ) 
	{ learner.dig[p][q]++; } else { digramic_overflow_warning = 1; }
      p = q;
      tok++;
    }

    if( digramic_overflow_warning ) {
      fprintf(stderr, 
	      "warning: ran out of integers (too much data), "
	      "reference measure may be skewed.\n");
    }
    if( overflow_warning ) {
      fprintf(stderr, 
	      "warning: ran out of integers (too much data), "
	      "results may be skewed.\n");
    }
  }
}

/* initialize global learner object */
void init_learner() {
  alphabet_size_t i, j;
  regex_count_t c;
  token_order_t z;

  /* some constants */
  learner.max_tokens = default_max_tokens;
  learner.max_hash_bits = default_max_hash_bits;
  learner.full_token_count = 0;
  learner.unique_token_count = 0;
  learner.logZ = 0.0;
  learner.max_order = 0;

  /* init character frequencies */
  for(i = 0; i < ASIZE; i++) { 
    for(j = 0; j < ASIZE; j++) { 
      learner.dig[i][j] = 0.0; 
    }
  }

  /* open temporary file for writing tokens */
  learner.tmp = tmpfile();

  /* allocate room for hash */
  learner.hash = calloc(learner.max_tokens, sizeof(l_item));
  if( !learner.hash ) {
    fprintf(stderr, 
	    "error: not enough memory? I couldn't allocate %li bytes\n", 
	    (sizeof(l_item) * ((long int)learner.max_tokens)));
    exit(0);
  }

  for(c = 0; c < MAX_RE + 1; c++) {
    learner.regex_token_count[c] = 0;
  }

  for(z = 0; z < MAX_SUBMATCH; z++) {
    learner.fixed_order_token_count[z] = 0;
    learner.fixed_order_unique_token_count[z] = 0;
  }
}

/* initializes to the Laplacian parameters */ 
void init_dirichlet() {
    alphabet_size_t i;

    dirichlet.alpha = 1.0;
    for(i = 0; i < ASIZE; i++) {
	dirichlet.u[i] = 1.0/((double)ASIZE);
    }
}

/* calculates the most probable Dirichlet parameters 
   given digram counts. Method described in MacKay and Peto (1995)
   Since we don't count the DIAMOND-DIAMOND digram, including the prior will 
   overestimate the DIAMOND-DIAMOND transition probability. However, for this 
   transition as well as the others, the Dirichlet will have practically
   no effect when there's enough data in the corpus.
*/
void optimize_dirichlet() {
  alphabet_size_t i, j;
  token_count_t k;
  float G[ASIZE];
  float H[ASIZE];
  float V[ASIZE];
  float F[ASIZE];
  double prev_alpha;
  double K, t;

  /* precompute useful constants */
  for(j = 1; j < ASIZE; j++) {
    G[j] = 0.0;
    H[j] = 0.0;
    V[j] = 0.0;
    F[j] = 0.0;
    for(i = 1; i < ASIZE; i++) {
      for(k = 1; k < learner.dig[i][j]; k++) {
	G[j] += 1.0/k;
	H[j] += 1.0/((double)k*k);
      }
      if( learner.dig[i][j] > 0 ) { 
	V[j]++; 
	F[j] += learner.dig[i][j];
      }
    }
    if(0 && options & (1<<OPTION_DEBUG) ) { 
      fprintf(stdout, 
	      "symbol %d has G = %g H = %g V = %g F = %g\n", 
	      j, G[j], H[j], V[j], F[j]);
    }
  }

  /* now optimize the dirichlet parameters */
  do {
    prev_alpha = dirichlet.alpha;
    K = 0.0;
    for(i = 1; i < ASIZE; i++) {
      K += log((F[i] + prev_alpha)/prev_alpha) +
	(0.5) * F[i]/(prev_alpha * (F[i] + prev_alpha));
    }
    dirichlet.alpha = 0.0;
    for(j = 1; j < ASIZE; j++) {
      t = K - G[j];
      dirichlet.u[j] = 2 * V[j] /(t + sqrt(t * t + 4 * H[j] * V[j]));
      dirichlet.alpha += dirichlet.u[j];
    }
    if( options & (1<<OPTION_VERBOSE) ) {
      fprintf(stdout, 
	      "dirichlet alpha %g --> %g\n", 
	      prev_alpha, dirichlet.alpha);
    }
  } while( fabs(prev_alpha - dirichlet.alpha) > TOL );
  
  if(0 && options & (1<<OPTION_DEBUG) ) {
    for(j = 1; j < ASIZE; j++) {
      fprintf(stdout, 
	      "dirichlet u[%d] =  %g\n", 
	      j, dirichlet.u[j]);
    }
  }
}

/* computes the _logarithm_ of digram transition probabilities */
void compute_digram_probabilities() {
  alphabet_size_t i,j;
  weight_t t;
    
  for(i = 1; i < ASIZE; i++) {
    t = 0.0;
    for(j = 1; j < ASIZE; j++) {
      t += learner.dig[i][j];
    }
    for(j = 1; j < ASIZE; j++) {
      learner.dig[i][j] = 
	log(((double)learner.dig[i][j] + dirichlet.u[j]) / 
	    (t + dirichlet.alpha));
      /* simulate the effect of digitizing the digrams */
      learner.dig[i][j] = UNPACK_DIGRAMS(PACK_DIGRAMS(learner.dig[i][j])); 
      if(0 && options & (1<<OPTION_DEBUG) ) {
	fprintf(stdout,
		"learner.dig[%d][%d] = %f\n", 
		i, j, learner.dig[i][j]);
      }
    }
  }
}

weight_t calc_learner_digramic_excursion(char *tok) {
  register alphabet_size_t p, q;
  register weight_t t = 0.0;

  /* now update digram frequency counts */
  p = (unsigned char)*tok++;
  /* finally update character frequencies */
  while( *tok ) {
    q = (unsigned char)*tok;
    t += learner.dig[p][q];
    p = q;
    tok++;
  }
  return t;
}

/* calculates the rth-order divergence but needs normalizing constant 
   note: this isn't the full divergence from digref, just the bits needed 
   for the r-th optimization */
score_t learner_divergence(score_t z, token_order_t r) {
  register hash_count_t i;
  register token_count_t c = 0;
  score_t t = 0.0;

  for(i = 0; i < learner.max_tokens; i++) {
    if( FILLEDP(&learner.hash[i]) &&
	(learner.hash[i].order == r) ) {

      t += (learner.hash[i].lam) * learner.hash[i].count;

      if( ++c >= learner.fixed_order_unique_token_count[r] ) {
	c = 0;
	break;
      }
    }
  }

  return -log(z)/((weight_t)r) + 
    t/((weight_t)learner.fixed_order_token_count[r]);
}

/* calculates the normalizing constant */
score_t learner_Z(token_order_t r, score_t unchanging_part) {
  register hash_count_t i;
  register token_count_t c = 0;

  score_t t = 0.0;

  for(i = 0; i < learner.max_tokens; i++) {
    if( FILLEDP(&learner.hash[i]) && 
	(learner.hash[i].order == r) ) {

      t += (exp(r * (learner.hash[i].lam)) - 1.0) * 
	exp(r * UNPACK_LWEIGHTS(learner.hash[i].ltrms) +
	    UNPACK_LWEIGHTS(learner.hash[i].dref));

      if( ++c >= learner.fixed_order_unique_token_count[r] ) {
	c = 0;
	break;
      }
    }
  }
  /* investigate this someday */
  if( isnan(t + unchanging_part) ) {
    fprintf(stderr, "error: sorry, partition function went kaboom.\n");
    exit(0);
  }
  //printf("z = %" FMT_printf_score_t "\n", t + unchanging_part);
  return t + unchanging_part;
}

/* fills the hash with partial calculations */
void fill_ref_vars(l_item *k, char *tok, token_order_t r) {
  hash_value_t id;
  char *t;
  l_item *l;

  /* weight of the r-th order excursion */
  k->dref = PACK_LWEIGHTS(calc_learner_digramic_excursion(tok));

  /* for each prefix of tok, add its weight */
  k->ltrms = PACK_LWEIGHTS(0.0);
  for( t = tok + strlen(tok) - 1; t > tok + 1; t-- ) {
    if( *(t - 1) == DIAMOND ) {
      *t = '\0';
      id = (hash_value_t)hash(tok, t - tok, 0);
      l = find_in_learner(id); /* guaranteed to be found */
      k->ltrms += PACK_LWEIGHTS(l->lam);
    }
  }

}

/* fills hash with partial calculations and returns the
   unchanging part of the normalizing constant */
score_t recalculate_reference_measure(token_order_t r) {
  hash_value_t id;
  char buf[BUFLEN];
  char tok[MAX_TOKEN_LEN];

  char *p, *q;
  l_item *k;

  score_t partial_z = 1.0;

  /* now we calculate the logarithmic word weight
     from the digram model, for each token in the hash */
  rewind(learner.tmp);
  q = tok;
  while( !feof(learner.tmp) && fgets(buf, BUFLEN, learner.tmp) ) {
    p = buf;
    while( *p ) {
      if( *p != -1) {
	*q++ = *p; /* copy into tok */
      } else { /* interword space */ 
	*q = 0; /* append NUL to tok */
	/* now write weight in hash */
	id = (hash_value_t)hash(tok, strlen(tok), 0);
	k = find_in_learner(id); /* guaranteed to be found */
	fill_ref_vars(k, tok, r);

	/* now calculate the unchanging part of the normalizing constant */
	if(k->order < r) {
	  partial_z += (exp(r * (k->lam)) - 1.0) * 
	    exp(r * UNPACK_LWEIGHTS(k->ltrms) + 
		UNPACK_LWEIGHTS(k->dref));
	}
	q = tok; /* reset q */
      }
      p++;
    }
  }
  /* finish last word */
  *q = 0; /* append NUL to tok */
  /* now write weight in hash */
  id = (hash_value_t)hash(tok, strlen(tok), 0);
  k = find_in_learner(id); /* guaranteed to be found */
  fill_ref_vars(k, tok, r);	
  /* now calculate the unchanging part of the normalizing constant */
  if(k->order < r) {
    partial_z += (exp(r * (k->lam)) - 1.0) * 
      exp(r * UNPACK_LWEIGHTS(k->ltrms) + 
	  UNPACK_LWEIGHTS(k->dref));
  }

  /* done */

  return partial_z;
}

/* just for debugging */
void print_score(token_order_t r) {
  hash_count_t i;
  double logprob = 0.0;
  double lpapprox = 0.0;
  double norm;

  for(i = 0; i < learner.max_tokens; i++) {
    if( FILLEDP(&learner.hash[i]) &&
	(learner.hash[i].order <= r) ) {
      logprob += learner.hash[i].count * (learner.hash[i].lam);
      if( learner.hash[i].order == 1 ) {
	logprob += learner.hash[i].count * 
	  UNPACK_LWEIGHTS(learner.hash[i].dref)/((weight_t)r);
      }
    }
    if( FILLEDP(&learner.hash[i]) &&
	(learner.hash[i].order == r) ) {
      lpapprox += learner.hash[i].count * 
	((learner.hash[i].lam) + 
	 UNPACK_LWEIGHTS(learner.hash[i].ltrms) + 
	 UNPACK_LWEIGHTS(learner.hash[i].dref)/((weight_t)r));
    }

  }
  norm = learner.fixed_order_token_count[r] * learner.logZ;

  printf("*** logprob = %" FMT_printf_score_t " * %d (r = %d, logZ = %" FMT_printf_score_t ")\n", 
	 ((score_t)(logprob / learner.fixed_order_token_count[r])), 
	 learner.fixed_order_token_count[r], 
	 r, learner.logZ);
  printf("*** lpapprox = %" FMT_printf_score_t " * %d (r = %d, logZ = %" FMT_printf_score_t ")\n", 
	 ((score_t)(lpapprox / learner.fixed_order_token_count[r])), 
	 learner.fixed_order_token_count[r], 
	 r, learner.logZ);
}

/* minimizes the divergence by solving for lambda one 
   component at a time.  */
void minimize_learner_divergence() {
  hash_count_t i;
  token_order_t r;
  token_count_t c = 0;
  score_t d, dd, z;
  score_t thresh, old_lam;
  score_t upz, div_extra_bits;

  if( options & (1<<OPTION_VERBOSE) ) {
    fprintf(stdout, "now maximizing model entropy\n");
  }

  div_extra_bits = 0.0;
  learner.logZ = 0.0;
  learner.divergence = 0.0;

  for(r = 1; 
      r <= ((options & (1<<OPTION_MULTINOMIAL)) ? 1 : learner.max_order); 
      r++) {

    /* here we precalculate various bits and pieces
       which aren't going to change during this iteration */
    upz = recalculate_reference_measure(r);
    if( r > 1 ) {
      /* calculate extra bits for divergence score */
      z = 0.0;
      for(i = 0; i < learner.max_tokens; i++) {
	if( FILLEDP(&learner.hash[i]) &&
	    (learner.hash[i].order == (r - 1)) ) {
	  z += (learner.hash[i].lam) * learner.hash[i].count;
	}
      }
      div_extra_bits += (learner.fixed_order_token_count[r - 1] > 0) ? 
	(z/learner.fixed_order_token_count[r-1]) : 0;
    }

    if( options & (1<<OPTION_REFMODEL) ) {
      break;
    }

    if( options & (1<<OPTION_VERBOSE) ) {
      fprintf(stdout, "* optimizing order %d weights (%i tokens, %i unique)\n", 
	      r, learner.fixed_order_token_count[r],
	      learner.fixed_order_unique_token_count[r]);
    }

    z = learner_Z(r, upz);
    dd = learner_divergence(z, r);

    do {
      d = dd;     /* save old divergence */

      thresh = 0.0;
      /* change each lambda value one at a time,
	 since we only rarely recalculate Z,
	 we randomly decide to update a given lambda to
	 prevent large errors and periodic behaviour */
      for(i = 0; i < learner.max_tokens; i++) {
	if( FILLEDP(&learner.hash[i]) &&
	    (learner.hash[i].order == r) ) {
	  if( rand() > (RAND_MAX>>3)) {

	    old_lam = (learner.hash[i].lam);

#define FIRST_METHOD
#if defined FIRST_METHOD
	    learner.hash[i].lam = 
	      ( (log((learner.hash[i].count * z) / 
		     learner.fixed_order_token_count[r]) /
		 ((weight_t)r) ) -
		UNPACK_LWEIGHTS(learner.hash[i].ltrms) - 
		UNPACK_LWEIGHTS(learner.hash[i].dref) / 
		((weight_t)r) );
#else
	    learner.hash[i].lam = 
	      ( log((learner.hash[i].count * 
		     (z - 
		      exp(r * 
			  ((learner.hash[i].lam) + 
			   UNPACK_LWEIGHTS(learner.hash[i].ltrms)) +
			  UNPACK_LWEIGHTS(learner.hash[i].dref)) )) /
		    (learner.fixed_order_token_count[r] -
		     learner.hash[i].count) ) / 
		((weight_t)r) -
		UNPACK_LWEIGHTS(learner.hash[i].ltrms) -
		UNPACK_LWEIGHTS(learner.hash[i].dref) / 
		((weight_t)r) );
#endif

	    if( isnan((learner.hash[i].lam)) ) {
	      /* precision problem, just ignore */
	      learner.hash[i].lam = (old_lam);
	      z = learner_Z(r, upz);
	      thresh = 0.0;
	    } else {
	      /* this ad hoc bit prevents really big changes in lambda,
		 should help stability and prevent too many threshold 
		 recalculations of Z */
#define MAX_LAMBDA_JUMP 5
	      if( (learner.hash[i].lam) > 
		  (old_lam + MAX_LAMBDA_JUMP) ) {
		learner.hash[i].lam = (old_lam + MAX_LAMBDA_JUMP);
	      } else if( (learner.hash[i].lam) < 
			 (old_lam - MAX_LAMBDA_JUMP) ) {
		learner.hash[i].lam = (old_lam - MAX_LAMBDA_JUMP);
	      }

	      /* if the changes so far add up to a lot, we should recalculate Z */
	      thresh += fabs(old_lam - (learner.hash[i].lam)) * 
		learner.hash[i].count;
	    }

	  }
	  if( thresh > learner.fixed_order_token_count[r] ) {
	    z = learner_Z(r, upz);
	    thresh = 0.0;
	  }
	  if( ++c >= learner.fixed_order_unique_token_count[r] ) {
	    c = 0;
	    break;
	  }
	}
      }
      /* update values */
      z = learner_Z(r, upz);
      dd = learner_divergence(z, r);

      if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stdout, "entropy change %" FMT_printf_score_t \
	                " --> %" FMT_printf_score_t "\n", 
		d + div_extra_bits, 
		dd + div_extra_bits);
      }

    } while( fabs(d - dd) > TOL );

    learner.logZ = log(z)/((weight_t)r);
    learner.divergence = dd + div_extra_bits;
  }

}

/* dumps readable model weights to the output */
void dump_model(FILE *out, FILE *in) {

  hash_value_t id;
  char buf[BUFLEN];
  char tok[MAX_TOKEN_LEN];
  char smb[MAX_SUBMATCH+1];
  token_order_t s;
  regex_count_t c;
  char *p, *q;
  l_item *k;

  /* preamble - this is copied from save_learner */

  fprintf(out, MAGIC1, learner.filename, 
	  (options & (1<<OPTION_REFMODEL)) ? "(ref)" : "");
  fprintf(out, MAGIC2_o, learner.divergence, learner.logZ, learner.max_order,
	  (options & (1<<OPTION_MULTINOMIAL)) ? "multinomial" : "hierarchical" );
  fprintf(out, MAGIC3, learner.max_hash_bits, 
	  learner.full_token_count, learner.unique_token_count);

  /* print out any regexes we might need */
  for(c = 0; c < regex_count; c++) {
    /* write the bitmap */
    for(p = smb, s = 1; s <= MAX_SUBMATCH; s++) {
      if( re[c].submatches & (1<<s) ) {
	*p++ = s + '0';
      }
    }
    *p = '\0';

#if defined HAVE_LIBBOOST_REGEX
    /* does this work? */
    fprintf(out, MAGIC5_o, re[c].string, smb);
#else
    fprintf(out, MAGIC5_o, re[c].string, smb);
#endif
  }
  /* this is optional too */
  if( options & (1<<OPTION_CASEN) ) {
    fprintf(out, MAGIC4);
  }
  if( options & (1<<OPTION_I18N) ) {
    fprintf(out, MAGIC7);
  }
  fprintf(out, MAGIC6); 

  fprintf(out, "# lambda | dig_ref | count | token\n");


  /* now go through hash printing values */
  rewind(learner.tmp);
  q = tok;
  while( !feof(learner.tmp) && fgets(buf, BUFLEN, learner.tmp) ) {
    p = buf;
    while( *p ) {
      if( *p != -1) {
	*q++ = *p; /* copy into tok */
      } else { /* interword space */ 
	*q = 0; /* append NUL to tok */
	/* now write weight in hash */
	id = (hash_value_t)hash(tok, strlen(tok), 0);
	k = find_in_learner(id); /* guaranteed to be found */
        fprintf(out, "%9.3f %9.3f %7d %s\n", 
                (weight_t)k->lam, UNPACK_LWEIGHTS(k->dref), k->count, tok);
	q = tok; /* reset q */
      }
      p++;
    }
  }
  /* finish last word */
  *q = 0; /* append NUL to tok */
  /* now write weight in hash */
  id = (hash_value_t)hash(tok, strlen(tok), 0);
  k = find_in_learner(id); /* guaranteed to be found */
  fprintf(out, "%9.3f %9.3f %7d %s\n", 
	  (weight_t)k->lam, UNPACK_LWEIGHTS(k->dref), k->count, tok);

}

void optimize_learner_and_save() {

  hash_count_t i;
  token_order_t c;

  if(100 * learner.unique_token_count >= HASH_FULL * learner.max_tokens) { 
    fprintf(stderr, 
	    "warning: table full, some tokens ignored - "
	    "try with option -h %i\n",
	    learner.max_hash_bits + 1);
  } else if( learner.unique_token_count <= 0 ) {
    fprintf(stderr, 
	    "warning: no tokens matched - have I learned nothing?\n");
  }

  if( skewed_constraints_warning ) {
    fprintf(stderr,
	    "warning: ran out of integers (too much data) constraints will be skewed.\n");
  }

  if( options & (1<<OPTION_VERBOSE) ) {
    fprintf(stdout, 
	    "picked up %i (%i distinct) tokens\n", 
	    learner.full_token_count, learner.unique_token_count);
    fprintf(stdout, 
	    "calculating reference word weights\n");
  }

  /* find best Dirichlet parameters */
  init_dirichlet();
  optimize_dirichlet();

  /* estimate the conditional digram probabilities */
  compute_digram_probabilities();

  if( learner.fixed_order_token_count[1] == 0 ) {
    /* it's a higher order model but there are no first
       order tokens! We can't handle that! Go through the 
       hash converting everything to first order.
       This will result in incorrect calculations, unless
       the higher order tokens don't overlap. Suitable only for
       geniuses and fools. */
    fprintf(stderr, 
	    "\n"
	    "warning: you have not defined any unigrams, so in this model\n"
	    "         features will be treated independently, which is quite\n"
	    "         likely incorrect. I hope you know what you're doing,\n"
	    "         because I don't!\n\n");

    options |= (1<<OPTION_MULTINOMIAL);
    for(c = 2; c <= learner.max_order; c++) {
      learner.fixed_order_token_count[1] += learner.fixed_order_token_count[c];
      learner.fixed_order_unique_token_count[1] += 
	learner.fixed_order_unique_token_count[c];
    }
    for(i = 0; i < learner.max_tokens; i++) {
      if( FILLEDP(&learner.hash[i]) ) {
	learner.hash[i].order = 1;
      }
    }

  }

  /* policy: always treat order 1 models multinomially */
  if( learner.max_order == 1 ) {
    options |= (1<<OPTION_MULTINOMIAL);
  }

  minimize_learner_divergence();

  if( options & (1<<OPTION_DUMP) ) {
    rewind(learner.tmp);
    dump_model(stdout, learner.tmp);
  }
  fclose(learner.tmp);

  /* now save the model to a file */
  save_learner();
}

/***********************************************************
 * MULTIBYTE FILE HANDLING FUNCTIONS                       *
 * this is suitable for any locale whose character set     *
 * encoding doesn't include NUL bytes inside characters    *
 ***********************************************************/

/* returns true if the line should be processed further
   depends on global mbox state */
/* note: mail filtering is very complicated, since
   any message can be mangled and requoted by a sender
   this will only cope with clean mail boxes */
bool_t mbox_line_filter(char *line) {

  bool_t process_line = 0; /* by default we skip the line */

  /* below we decide if we want to process the line */
  switch(mbox.state) {
  case UNDEF:
    if( mbox.prev_line_empty && 
	(!strncmp(line, "From ", 5) || 
	 ((line[0] == '-') && (line[1] == '-') && !iswspace(line[2]))) ) {
      /* if it starts with From_ or else it looks like --xxxxx, 
	 it signals a new header */
      mbox.state = HEADER;
      mbox.checked_content_type = 0;
    }
    break;
  case HEADER:
    if( *line == '\n' ) {
      mbox.state = BODY;
    } else if( !strncmp(line, "Content-Type:", 13) ) {
      mbox.checked_content_type = 1;
      if( strstr(line + 13, "text/") || 
	  strstr(line + 13, "TEXT/") ) {
	mbox.mime_type = 1; /* good */
      } else {
	mbox.mime_type = 0; /* bad */
      }
    } else if( !strncmp(line, "Subject:", 8) ||
	       !(strncmp(line, "From:", 5)) ) {
      /* process subject and from line like body */
      process_line = 1;
    }
    break;
  case BODY:
      if( mbox.prev_line_empty && 
	  (!strncmp(line, "From ", 5) || 
	   ((line[0] == '-') && (line[1] == '-') && !isspace(line[2])))) {
	/* if it starts with From_ or else it looks like --xxxxx, 
	   it signals a new header */
	mbox.state = HEADER;
	mbox.checked_content_type = 0;
      } else if( mbox.checked_content_type && mbox.mime_type ) { 
	process_line = 1;
      } else if( !mbox.checked_content_type ) {
	process_line = 1; /* be lenient */
      }
      break;
    }
    mbox.prev_line_empty = (*line == '\n') ? 1 : 0; /* for next time */

    return process_line;
}

/* removes tags in the string - modifies in place */
/* the name of this function is a misnomer, since it doesn't
   parse xml properly. But we just want a simple kludge 
   for most html flavours */
void xml_character_filter(char *line) {

  char *q;

  q = line;
  while( *line ) {
    switch(xml.state) {
    case TEXT:
      /* does it look like <x where x is either alpha or punctuation? */
      if( line[0] == '<' ) {
	if( !strncmp(line + 1, "!--", 3) ) {
	  xml.state = COMMENT;
	  line += 3;
	} else if( !strncasecmp(line + 1, "script", 6) ) {
	  xml.state = SPECIAL;
	  line += 6;
	} else if( isalpha(line[1]) || ispunct(line[1]) ) {
	  xml.state = TAG;
	  line += 1;
	}
      } else {
	*q++ = *line;
      }
      break;
    case TAG:
      if( (line[1] == '>') && 
	  (isalpha(*line) || ispunct(*line)) ) {
	xml.state = TEXT;
	line++; 
      }
      break;
    case COMMENT:
      if( line[0] == '-' && !strncmp(line + 1, "->", 2) ) {
	xml.state = TEXT;
	line += 2;
      }
      break;
    case SPECIAL:
      if( line[0] == '<' ) {
	if( !strncasecmp(line + 1, "/script", 7) ) {
	  xml.state = TEXT;
	  line += 7;
	}
      }
      break;
    }
    line++;
  }
  *q = '\0'; /* mark the end of the clean text string */
}

#define MULTIBYTE_EPSILON 10 /* enough for a multibyte char and a null char */

/* reads a text file as input, converting each line
into a wide character representation and applies several
filters. */
void process_file(FILE *input, 
		  int (*line_filter)(char *),
		  void (*character_filter)(char *), 
		  void (*word_fun)(char *, token_order_t, regex_count_t), 
		  void (*line_fun)(void)) {
  char *s;
  regex_count_t i;
  charbuf_len_t k;
  int eflag;
  token_order_t z, j, n, order;
  charbuf_len_t l;

  char *p, *q, *qq;
  char tok[MAX_TOKEN_LEN+2];

  regmatch_t pmatch[MAX_SUBMATCH];

  char ntok[MAX_TOKEN_LEN+2];
  char *nq;
  token_order_t nhow_many;

#if defined HAVE_LIBBOOST_REGEX
  charbuf_len_t wclen;
  wchar_t *wp;
  mbstate_t tok_shiftstate, input_shiftstate;
  charbuf_len_t tok_len;

  memset(&input_shiftstate, 0, sizeof(mbstate_t));
#endif


  /* initialize the norex state */
  ntok[0] = DIAMOND;
  ntok[1] = '\0';
  nq = ntok + 1;
  nhow_many = 0;
 
  /* now start processing */

  while( !feof(input) ) {
    /* read in a full line, allocating memory as necessary */
    textbuf[0] = '\0';
    s = textbuf;
    l = textbuf_len;
    k = 1;
    while( fgets(s, l, input) && (strlen(s) >= (l - 1)) ) {
      textbuf = realloc(textbuf, 2 * textbuf_len);
      if( !textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for input line (%d bytes)\n",
		textbuf_len);
	exit(0);
      }

      s = textbuf + textbuf_len - (k++);
      l = textbuf_len;
      textbuf_len *= 2;
    }

    /* if appending, print the lines as they 
       come in before they are processed */
    if( options & (1<<OPTION_APPEND) ) {
      fprintf(stdout, " %s", textbuf);
    }

    /* next we check to see if this line should be skipped */
    if( *textbuf && (!line_filter || (*line_filter)(textbuf)) ) {
      /* now filter some of the characters in the current line */
      if( character_filter ) { (*character_filter)(textbuf); }

      /* repeat for each regular expression:
	 find all the instances of a matching substring */
#if defined HAVE_LIBBOOST_REGEX

      /* damn - we compiled wide character regexes, 
	 now we've got to use them (not really, but who in their right
	 mind would compile with boost if they're not going to convert
	 to wide characters anyway) */

      if( textbuf_len > wc_textbuf_len ) {
	wc_textbuf_len = textbuf_len;
	wc_textbuf = realloc(wc_textbuf, wc_textbuf_len * sizeof(wchar_t));
	if( !wc_textbuf ) {
	  fprintf(stderr, 
		  "error: not enough memory for wide character conversion "
		  "(%d bytes)\n",
		  wc_textbuf_len * sizeof(wchar_t));
	  exit(0);
	}
      }

      /* convert as much as we can of the line into wide characters */
      s = textbuf;
      k = textbuf_len;
      wp = wc_textbuf;
      wclen = 0;
      /* since we ensured textbuf_len <= wctextbuf_len
	 there will never be overflow of wctextbuf below */
      while( k > 0 ) {
	l = mbrtowc(wp, s, k, &input_shiftstate);
	if( l > 0 ) {
	  wp++;
	  wclen++;
	  k -= l;
	  s += l;
	} else if( l == 0 ) {
	  break;
	} else if( l == -1 ) {
	  /* try to be robust */
	  s++; 
	  k--;
	  memset(&input_shiftstate, 0, sizeof(mbstate_t));
	} else if( l == -2) {
	  /* couldn't parse a complete character */
	  break;
	}
      }
      *wp = L'\0';

      /* now process the regex */
      for(i = 0; i < regex_count; i++) {
	k = 0;
	eflag = 0;
	/* see if a match */
	while( (k < wclen) && (regexec(&re[i].regex, wc_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get converted, concatenated and the result gets word_fun'd */
	  q = tok;
	  *q++ = DIAMOND;
	  memset(&tok_shiftstate, 0, sizeof(mbstate_t));
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
		if( options & (1<<OPTION_CASEN) ) {
		  tok_len = wcrtomb(q, wc_textbuf[k + j], &tok_shiftstate); 
		} else {
		  tok_len = wcrtomb(q, towlower(wc_textbuf[k + j]), 
				    &tok_shiftstate); 
		}
		if( (tok_len > -1) ) {
		  q += tok_len;
		}
	      }
	    }
	    *q++ = DIAMOND;
	  }
 	  *q = '\0'; 

 	  /* now let each category process the token */
	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
	}
      }
#else 
      /* the GNU regex routines expect an ordinary string */
      for(i = 0; i < regex_count; i++) {
	k = 0;
 	eflag = 0;
 	/* see if a match */
 	while( (k < l) && (regexec(&re[i].regex, textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get concatenated and the result gets word_fun'd */
 	  q = tok;
	  *q++ = DIAMOND;
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN ) {
		if( options & (1<<OPTION_CASEN) ) {
		  *q++ = textbuf[k + j];
		} else {
		  *q++ = tolower(textbuf[k + j]);
		}
 	      }
 	    }
	    *q++ = DIAMOND;
 	  }
 	  *q = '\0'; 
 	  /* now let each category process the token */
	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
 	}	      
      }
      
#endif

      /* default processing: reads tokens and passes them to
	 the word_fun */
      if( options & (1<<OPTION_NOREGEX) ) {
	p = textbuf;

	while( *p ) {
	  if( isalpha(*p) ) {
	    if( nq < ntok + MAX_TOKEN_LEN ) {
	      if( !(options & (1<<OPTION_CASEN)) ) {
		*nq++ = tolower(*p);
	      } else {
		*nq++ = *p;
	      }
	    }
	  } else if( *(nq - 1) != DIAMOND ) { /* token boundary */
	    *nq++ = DIAMOND;
	    *nq = '\0';
	    if( ngram_order == 1 ) {
	      /* let each category process the token */
	      (*word_fun)(ntok, (token_order_t)1, 0);
	      nq = ntok;
	      *nq++ = DIAMOND;
	    } else {
	      if( ++nhow_many > ngram_order ) {
		nhow_many--;
		/* move all tokens down by one */
		for(nq = ntok + 1; *nq != DIAMOND; nq++) {};
		for(nq++, qq = ntok + 1; *nq; *qq++ = *nq++) {};
		*qq = '\0';
		nq = qq;
	      }

	      qq = ntok;
	      for(n = nhow_many; n > 0; n--) {
		/* let each category process the token */
		(*word_fun)(qq, (token_order_t)n, 0);
		qq++;
		/* skip to next token and repeat */
		while(*qq != DIAMOND ) { qq++; }
	      }
	    }
	  }
	  p++;
	}
      }

      /* now summarize this line if required */
      if( line_fun ) { (*line_fun)(); }
      
    } else { 
      /* since we don't process this line, we
	 reset XML tag state for next time */
      xml.state = TEXT; 
    }
  }
}

/***********************************************************
 * WIDE CHARACTER FILE HANDLING FUNCTIONS                  *
 * this is needed for any locale whose character set       *
 * encoding can include NUL bytes inside characters        *
 ***********************************************************/

/* returns true if the line should be processed further
   depends on global mbox state */
bool_t w_mbox_line_filter(wchar_t *line) {

  bool_t process_line = 0; /* by default we skip the line */

  /* below we decide if we want to process the line */
  switch(mbox.state) {
  case UNDEF:
    if( mbox.prev_line_empty && 
	(!wcsncmp(line, L"From ", 5) || 
	 ((line[0] == L'-') && (line[1] == L'-') && !iswspace(line[2]))) ) {
      /* if it starts with From_ or else it looks like --xxxxx, 
	 it signals a new header */
      mbox.state = HEADER;
      mbox.checked_content_type = 0;
    }
    break;
  case HEADER:
    if( *line == L'\n' ) {
      mbox.state = BODY;
    } else if( !wcsncmp(line, L"Content-Type:", 13) ) {
      mbox.checked_content_type = 1;
      if( wcsstr(line + 13, L"text/") || 
	  wcsstr(line + 13, L"TEXT/") ) {
	mbox.mime_type = 1; /* good */
      } else {
	mbox.mime_type = 0; /* bad */
      }
    } else if( !wcsncmp(line, L"Subject:", 8) ||
	       !(wcsncmp(line, L"From:", 5)) ) {
      /* process subject and from line like body */
      process_line = 1;
    }
    break;
  case BODY:
    if( mbox.prev_line_empty && 
	(!wcsncmp(line, L"From ", 5) || 
	 ((line[0] == L'-') && (line[1] == L'-') && !iswspace(line[2])))) {
      /* if it starts with From_ or else it looks like --xxxxx, 
	 it signals a new header */
      mbox.state = HEADER;
      mbox.checked_content_type = 0;
    } else if( mbox.checked_content_type && mbox.mime_type ) { 
      process_line = 1;
    } else if( !mbox.checked_content_type ) {
      /* unless we explicitly are told it's not text, we process it */
      /* this will let quoted uuencoded content through :-( */
      process_line = 1;
    }
    break;
  }
  mbox.prev_line_empty = (*line == L'\n') ? 1 : 0; /* for next time */

  return process_line;
}

/* removes tags in the string - modifies in place */
/* the name of this function is a misnomer, since it doesn't
   parse xml properly. But we just want a simple kludge 
   for most html flavours */
void w_xml_character_filter(wchar_t *line) {

  wchar_t *q;

  q = line;
  while( *line ) {
    switch(xml.state) {
    case TEXT:
      /* does it look like <x where x is either alpha or punctuation? */
      if( line[0] == L'<' ) {
	if( !wcsncmp(line + 1, L"!--", 3) ) {
	  xml.state = COMMENT;
	  line += 3;
	} else if( wcsncasecmp(line + 1, L"script", 6) != 0 ) {
	  xml.state = SPECIAL;
	  line += 6;
	} else if( iswalpha(line[1]) || iswpunct(line[1]) ) {
	  xml.state = TAG;
	  line += 1;
	}
      } else {
	*q++ = *line;
      }
      break;
    case TAG:
      if( (line[1] == L'>') && 
	  (iswalpha(*line) || iswpunct(*line)) ) {
	xml.state = TEXT;
	line++; 
      }
      break;
    case COMMENT:
      if( line[0] == L'-' && !wcsncmp(line + 1, L"->", 2) ) {
	xml.state = TEXT;
	line += 2;
      }
      break;
    case SPECIAL:
      if( line[0] == L'<' ) {
	if( !wcsncasecmp(line + 1, L"/script", 7) ) {
	  xml.state = TEXT;
	  line += 7;
	}
      }
      break;
    }
    line++;
  }
  *q = L'\0'; /* mark the end of the clean text string */
}

/* reads a text file as input, converting each line
into a wide character representation and applies several
filters. */
void w_process_file(FILE *input, 
		    int (*line_filter)(wchar_t *),
		    void (*character_filter)(wchar_t *), 
		    void (*word_fun)(char *, token_order_t, regex_count_t), 
		    void (*line_fun)(void)) {
  char *s;
  regex_count_t i;
  charbuf_len_t k;
  int eflag;
  token_order_t z, j, n, order;
  charbuf_len_t l;
  charbuf_len_t wclen;

  wchar_t *wp;
  mbstate_t tok_shiftstate, input_shiftstate;
  charbuf_len_t tok_len;
  char *q, *qq;
  char tok[MAX_TOKEN_LEN+2];

  regmatch_t pmatch[MAX_SUBMATCH];

  char *nq;
  char ntok[MAX_TOKEN_LEN+2];
  token_order_t nhow_many;

  /* initialize the norex state */
  ntok[0] = DIAMOND;
  ntok[1] = '\0';
  nq = ntok + 1;
  nhow_many = 0;

  memset(&input_shiftstate, 0, sizeof(mbstate_t));

  while( !feof(input) ) {

    /* read in a full line, allocating memory as necessary */
    textbuf[0] = '\0';
    s = textbuf;
    l = textbuf_len;
    k = 1;
    while( fgets(s, l, input) && (strlen(s) >= (l - 1)) ) {
      textbuf = realloc(textbuf, 2 * textbuf_len);
      if( !textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for input line (%d bytes)\n",
		textbuf_len);
	exit(0);
      }

      s = textbuf + textbuf_len - (k++);
      l = textbuf_len;
      textbuf_len *= 2;
    }

    /* if appending, print the lines as they 
       come in before they are processed */
    if( options & (1<<OPTION_APPEND) ) {
      fprintf(stdout, " %s", textbuf);
    }

    /* now convert the line into a wide character string */
    if( textbuf_len > wc_textbuf_len ) {
      wc_textbuf_len = textbuf_len;
      wc_textbuf = realloc(wc_textbuf, wc_textbuf_len * sizeof(wchar_t));
      if( !wc_textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for wide character conversion "
		"(%d bytes)\n",
		wc_textbuf_len * sizeof(wchar_t));
	exit(0);
      }
    }

    /* convert as much as we can of the line into wide characters */
    s = textbuf;
    k = textbuf_len;
    wp = wc_textbuf;
    wclen = 0;
    /* since we ensured textbuf_len <= wctextbuf_len
       there will never be overflow of wctextbuf below */
    while( k > 0 ) {
      l = mbrtowc(wp, s, k, &input_shiftstate);
      if( l > 0 ) {
	wp++;
	wclen++;
	k -= l;
	s += l;
      } else if( l == 0 ) {
	break;
      } else if( l == -1 ) {
	/* try to be robust */
	s++; 
	k--;
	memset(&input_shiftstate, 0, sizeof(mbstate_t));
      } else if( l == -2) {
	/* couldn't parse a complete character */
	break;
      }
    }
    *wp = L'\0';

    /* next we check to see if this line should be skipped */
    if( (wclen > 0) && 
	(!line_filter || (*line_filter)(wc_textbuf)) ) {
      /* now filter some of the characters in the current line */
      if( character_filter ) { (*character_filter)(wc_textbuf); }

      /* repeat for each regular expression:
	 find all the instances of a matching substring */
#if defined HAVE_LIBBOOST_REGEX

      for(i = 0; i < regex_count; i++) {
	k = 0;
	eflag = 0;
	/* see if a match */
	while( (k < wclen) && (regexec(&re[i].regex, wc_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get converted, concatenated and the result gets word_fun'd */
	  q = tok;
	  *q++ = DIAMOND;
	  memset(&tok_shiftstate, 0, sizeof(mbstate_t));
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
		if( options & (1<<OPTION_CASEN) ) {
		  tok_len = wcrtomb(q, wc_textbuf[k + j], &tok_shiftstate); 
		} else {
		  tok_len = wcrtomb(q, towlower(wc_textbuf[k + j]), 
				    &tok_shiftstate); 
		}
		if( (tok_len > -1) ) {
		  q += tok_len;
		}
	      }
	    }
	    *q++ = DIAMOND;
	  }
 	  *q = '\0'; 

 	  /* now let each category process the token */
 	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
	}
      }
#else 
      /* the GNU regex routines expect a multibyte string */
      if( textbuf_len > aux_textbuf_len ) {
	aux_textbuf_len = textbuf_len;
	aux_textbuf = realloc(aux_textbuf, aux_textbuf_len);
	if( !aux_textbuf ) {
	  fprintf(stderr, 
		  "error: not enough memory for auxiliary text buffer "
		  "(%d bytes)\n",
		  aux_textbuf_len);
	  exit(0);
	}
      }
      l = wcstombs(aux_textbuf, wc_textbuf, aux_textbuf_len - 1);
      aux_textbuf[aux_textbuf_len - 1] = '\0';

      for(i = 0; i < regex_count; i++) {
	k = 0;
 	eflag = 0;
 	/* see if a match */
 	while( (k < l) && (regexec(&re[i].regex, aux_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get concatenated and the result gets word_fun'd */
 	  q = tok;
	  *q++ = DIAMOND;
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN ) {
		if( options & (1<<OPTION_CASEN) ) {
		  *q++ = aux_textbuf[k + j];
		} else {
		  *q++ = tolower(aux_textbuf[k + j]);
		}
 	      }
 	    }
	    *q++ = DIAMOND; /* hope DIAMOND can never be matched inside regex */
 	  }
 	  *q = '\0'; 
 	  /* now let each category process the token */
	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
 	}	      
      }
      
#endif

      /* default processing: reads tokens, converting them
	 to multibyte representation before passing them to
	 the word_fun */
      if( options & (1<<OPTION_NOREGEX) ) {
	wp = wc_textbuf;
	memset(&tok_shiftstate, 0, sizeof(mbstate_t));

	while( *wp ) {
	  if( iswalpha(*wp) ) {
	    if( !(options & (1<<OPTION_CASEN)) ) {
	      *wp = towlower(*wp);
	    }

	    if( nq < ntok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
	      tok_len = wcrtomb(nq, *wp, &tok_shiftstate); 
	      if( (tok_len > -1) ) {
		nq += tok_len;
	      }
	    }
	  } else if( *(nq - 1) != DIAMOND ) { /* token boundary */
	    *nq++ = DIAMOND;
	    *nq = '\0';
	    if( ngram_order == 1 ) {
	      /* let each category process the token */
	      (*word_fun)(ntok, (token_order_t)1, 0);
	      nq = ntok;
	      *nq++ = DIAMOND;
	    } else {
	      if( ++nhow_many > ngram_order ) {
		nhow_many--;
		/* move all tokens down by one */
		for(nq = ntok + 1; *nq != DIAMOND; nq++) {};
		for(nq++, qq = ntok + 1; *nq; *qq++ = *nq++) {};
		*qq = '\0';
		nq = qq;
	      }

	      qq = ntok;
	      for(n = nhow_many; n > 0; n--) {
		/* let each category process the token */
		(*word_fun)(qq, (token_order_t)n, 0);
		qq++;
		/* skip to next token and repeat */
		while(*qq != DIAMOND ) { qq++; }
	      }
	    }
	  }
	  wp++;
	}
      }

      /* now summarize this line if required */
      if( line_fun ) { (*line_fun)(); }
      
    } else { 
      /* since we don't process this line, we
	 reset XML tag state for next time */
      xml.state = TEXT; 
    }

  }
}

/***********************************************************
 * MAIN FUNCTIONS                                          *
 ***********************************************************/

char *sanitize_path(char *in) {
  char *q;
  char *path;
  charbuf_len_t l;

  /* this bit likely fails in DOS ;-) */
  if( (*in != '/') && (*in != '.') && (path = getenv("DBACL_PATH")) ) {
    l = strlen(path);
    q = malloc(l + strlen(in) + 3);
    strcpy(q, path);
    if( q[l - 1] != '/' ) {
      q[l] = '/'; 
      q[l + 1] = 0;
    }
    strcat(q, in);
    return q;
  } else {
    return in;
  }
}

int main(int argc, char **argv) {

  category_count_t c;
  regex_count_t k;
  FILE *input;
  char op;
  char *p;

#if defined HAVE_LIBBOOST_REGEX
  wchar_t *rstring;
#endif

  void (*preprocess_fun)(void) = NULL;
  void (*word_fun)(char *, token_order_t, regex_count_t) = NULL;
  void (*line_fun)(void) = NULL;
  void (*postprocess_fun)(void) = NULL;

  /* set up internationalization */
  if( !setlocale(LC_ALL, "") ) {
    fprintf(stderr, 
	    "warning: could not set locale, internationalization disabled\n");
  } else {
    if( options & (1<<OPTION_VERBOSE) ) {
      fprintf(stderr, 
	      "warning: international locales not supported\n");
    }
  }

  /* parse the options */
  while( (op = getopt(argc, argv, "HDRVMNadivnrh:c:l:T:f:g:j:w:x:")) > -1 ) {

    switch(op) {
    case 'a':
      options |= (1<<OPTION_APPEND);
      break;
    case 'D':
      options |= (1<<OPTION_DEBUG);
      break;
    case 'M':
      options |= (1<<OPTION_MULTINOMIAL);
      break;
    case 'N':
      options |= (1<<OPTION_POSTERIOR);
      break;
    case 'R':
      if( cat_count >= MAX_CAT ) {
	fprintf(stderr, 
		"warning: maximum reached, random text category omitted\n");
      } else if( options & (1<<OPTION_LEARN) ) {
	fprintf(stderr, 
		"error: cannot use options -l and -R together\n");
	usage(argv);
	exit(0);
      } else {
	options |= (1<<OPTION_CLASSIFY);

	cat[cat_count].filename = "random";

	init_category(&cat[cat_count]);
	init_purely_random_text_category(&cat[cat_count]);

	cat_count++;
      }
      break;
    case 'T':
      if( !strncasecmp(optarg, "email", 5) ) {
	options |= (1<<OPTION_MBOX_FORMAT);
	options |= (1<<OPTION_XML); /* filter out HTML from messages */
      } else if( !strncasecmp(optarg, "xml", 3) ) {
	options |= (1<<OPTION_XML);
      } else { /* default */
	options |= (1<<OPTION_TEXT_FORMAT);
      }
      break;
    case 'V':
      fprintf(stdout, "dbacl version %s\n", VERSION);
#if defined HAVE_LIBBOOST_REGEX
      fprintf(stdout, "Using BOOST wide character modern regexes.\n");
#else
      fprintf(stdout, "Using GNU modern regexes.\n");
#endif
      fprintf(stdout, "Feature memory requirements: %d bytes (classifying), %d bytes (learning)\n", 
	      sizeof(c_item), sizeof(l_item));
      exit(1);
      break;
    case 'd':
      options |= (1<<OPTION_DUMP);
      break;
    case 'h': /* select memory size in powers of 2 */
      default_max_hash_bits = strtod(optarg, NULL);
      if( default_max_hash_bits > MAX_HASH_BITS ) {
	fprintf(stderr, 
		"warning: maximum hash size will be 2^%d\n", 
		MAX_HASH_BITS);
	default_max_hash_bits = MAX_HASH_BITS;
      }
      default_max_tokens = (1<<default_max_hash_bits);
      break;
    case 'j':
      options |= (1<<OPTION_CASEN);
      break;
    case 'n':
      options |= (1<<OPTION_SCORES);
      break;
    case 'c':
      if( cat_count >= MAX_CAT ) {
	fprintf(stderr, 
		"warning: maximum reached, category ignored\n");
      } else if( options & (1<<OPTION_LEARN) ) {
	fprintf(stderr, 
		"error: cannot use options -l and -c together\n");
	usage(argv);
	exit(0);
      } else {
	options |= (1<<OPTION_CLASSIFY);

	cat[cat_count].filename = sanitize_path(optarg);

	if( !*optarg ) {
	  fprintf(stderr, "error: category needs a name\n");
	  usage(argv);
	  exit(0);
	}
	if( !load_category(&cat[cat_count]) ) {
	  fprintf(stderr, 
		  "error: could not load category %s\n", 
		  cat[cat_count].filename);
	  exit(0);
	}
	ngram_order = (ngram_order < cat[cat_count].max_order) ? 
	  cat[cat_count].max_order : ngram_order;
	cat_count++;
      }
      break;
    case 'r':
      options |= (1<<OPTION_REFMODEL);
      break;
    case 'w':
      ngram_order = atoi(optarg);
      if( !*optarg || (ngram_order < 1) || (ngram_order > 9) ) {
	fprintf(stderr, 
		"error: the -w switch needs a number between 1 and 9\n");
	exit(0);
      }
      break;
    case 'x':
      decimation = atoi(optarg);
      if( !*optarg || (decimation < 1) || (decimation > MAX_HASH_BITS) ) {
	fprintf(stderr, 
		"warning: option -x ignored, needs an integer between 1 and %d\n", 
		MAX_HASH_BITS);
      } else {
	options |= (1<<OPTION_DECIMATE);
      }
      break;
    case 'f':
      if( filter_count >= MAX_CAT ) {
	fprintf(stderr, "warning: maximum reached, filter ignored\n");
      } else if( options & (1<<OPTION_LEARN) ) {
	fprintf(stderr, "error: cannot use options -l and -f together\n");
	usage(argv);
	exit(0);
      } else if( !*optarg ) {
	fprintf(stderr, "error: filter must be category name or number\n");
	usage(argv);
	exit(0);
      } else {
	options |= (1<<OPTION_FILTER);
	filter[filter_count] = -1;

	/* see if it's a known category */
	for(c = 0; c < cat_count; c++) {
	  if( !strcmp(cat[c].filename, optarg) ) {
	    filter[filter_count] = c;
	    break;
	  }
	}
	/* if not recognized, see if it's a number */
	if( filter[filter_count] < 0 ) {
	  filter[filter_count] = strtod(optarg, NULL) - 1;
	}
	if( filter[filter_count] < 0 ) { /* still not recognized */
	  fprintf(stderr, 
		  "error: unrecognized category in -f option [%s]\n", 
		  optarg);
	  usage(argv);
	  exit(0);
	}
	filter_count++;
      }
      break;
    case 'v':
      options |= (1<<OPTION_VERBOSE);
      break;
    case 'l': 
      if( options & (1<<OPTION_CLASSIFY) ) {
	fprintf(stderr, 
		"error: cannot use options -l and -c together\n");
	usage(argv);
	exit(0);
      } else if( options & (1<<OPTION_LEARN) ) {
	fprintf(stderr, 
		"error: option -l can only occur once\n");
	exit(0);
      } else {
	options |= (1<<OPTION_LEARN);
	learner.filename = sanitize_path(optarg);
	if( !*learner.filename ) {
	  fprintf(stderr, 
		  "error: category needs a name\n");
	  usage(argv);
	  exit(0);
	}
      }
      break;
    case 'g':
      if( options & (1<<OPTION_CLASSIFY) ) {
	fprintf(stderr, "error: must use option -l together with -F\n");
	usage(argv);
	exit(0);
      } else {
	/* set up the submatch bitmap */
	re[regex_count].submatches |= 0;

	if( (p = strrchr(optarg, '|')) && ( *(--p) == '|') ) {
	  /* assume string ents in ||12345, use as bitmap */
	  *p = '\0';
	  for(p += 2; *p; p++) {
	    /* assume ascii number positions */
	    if( !isdigit(*p) || (*p > '9') || (*p < '1')) {
	      fprintf(stderr, 
		      "warning: could not decode bitmap for %s\n", optarg);
	    } else {
	      re[regex_count].submatches |= (1<<(*p - '0'));
	    }
	  }
	} else { /* no bitmap specified */
	  re[regex_count].submatches = ~0;
	}

#if defined HAVE_LIBBOOST_REGEX
	/* BOOST regexes must be wide */
	if( !(rstring = malloc((strlen(optarg) + 1) * sizeof(wchar_t))) ) {
	  fprintf(stderr, 
		  "warning: could not prepare regular expression '%s', ignored\n",
		  optarg);
	} else if( mbstowcs(rstring, optarg, strlen(optarg)) < 0 ) {
	  fprintf(stderr,
		  "warning: could not convert regular expression '%s', ignored\n",
		  optarg);
	} else {
	  rstring[strlen(optarg)] = L'\0';
	  if( regcomp(&re[regex_count].regex, rstring, REG_EXTENDED) != 0 ) {
	    fprintf(stderr, 
		    "warning: could not compile regular expression '%s', ignored\n", 
		    optarg);
	  } else {
	    re[regex_count].string = rstring;
	    regex_count++;
	  }
	}
#else
	/* GNU regexes use regular strings */
	if( regcomp(&re[regex_count].regex, optarg, REG_EXTENDED) != 0 ) {
	  fprintf(stderr, 
		  "warning: could not compile regular expression '%s', ignored\n", 
		  optarg);
	} else {
	  re[regex_count].string = optarg;
	  regex_count++;
	}
#endif
      }
      break;
    case 'i':
      options |= (1<<OPTION_I18N);
      if( !strcmp(nl_langinfo(CODESET), "UTF-8") ) {
	fprintf(stderr, "warning: you have UTF-8, so -i is not needed.\n");
      }
      break;
    default:
      break;
    }
  }

  /* end option processing */
    
  /* consistency checks */
  if( ((options>>OPTION_CLASSIFY) & 1) + 
      ((options>>OPTION_LEARN) & 1) != 1 ) {
    fprintf(stderr, 
	    "error: please use either -c or -l option\n");
    usage(argv);
    exit(0);
  }

  if( (options & (1<<OPTION_DECIMATE)) &&
      !(options & (1<<OPTION_LEARN)) ) {
    fprintf(stderr,
	    "warning: option -x ignored, applies only when learning\n");
    options &= ~(1<<OPTION_DECIMATE);
  }

  if( options & (1<<OPTION_DUMP) ) {
    if( options & (1<<OPTION_CLASSIFY) ) {
      options &= ~(1<<OPTION_DUMP); /* no sense */
    } else if( options & (1<<OPTION_VERBOSE) ) {
      options &= ~(1<<OPTION_VERBOSE); /* verbose writes garbage to stdout */
      options &= ~(1<<OPTION_DEBUG);
    }
  }

  if( ((options>>OPTION_TEXT_FORMAT) & 1) + 
      ((options>>OPTION_MBOX_FORMAT) & 1) > 1 ) {
    fprintf(stderr, 
	    "error: please use one of either -T text or -T email options\n");
    usage(argv);
    exit(0);
  }

  if( (options & (1<<OPTION_APPEND)) &&
      (options & (1<<OPTION_FILTER)) ) {
    options &= ~(1<<OPTION_APPEND);
    fprintf(stderr, 
	    "warning: disabling option -a, because it cannot be used with -f\n");
  }

  /* decide if we need some options */

  if( !regex_count ) {
    options |= (1<<OPTION_NOREGEX);
  }

  if( options & (1<<OPTION_MULTINOMIAL) ) { 
    if( cat_count == 1 ) {
      if( cat[0].model_type == simple ) {
	options |= (1<<OPTION_CALCENTROPY);
      }
    } else if( cat_count > 1 ) {
      for(c = 1; c < cat_count; c++) {
	if( cat[c].model_type == sequential ) { break; }
	if( cat[c].retype != cat[c-1].retype ) { break; }
      }
      if( c == cat_count ) {
	options |= (1<<OPTION_CALCENTROPY);
      }
    }
    if( !(options & (1<<OPTION_CALCENTROPY)) ) {
      fprintf(stderr,
	      "warning: -M switch ignored. "
	      "Not all categories support multinomial calculations\n");
    }
  }

  if( options & (1<<OPTION_I18N) ) {
#if defined HAVE_LIBBOOST_REGEX
#else
    fprintf(stderr, 
	    "warning: regexes operate in multibyte encoding.");
#endif	
  }

  /* set up callbacks */

  if( options & (1<<OPTION_CLASSIFY) ) {

    if( options & (1<<OPTION_CALCENTROPY) ) {
      init_empirical(); /* sets cached to zero */
    }

    if( cat_count == 1 ) {
      /* single category */

      preprocess_fun = NULL;
      word_fun = score_word;
      if( options & (1<<OPTION_FILTER) ) {
	empirical.track_features = 1; /* complex calc should be guaranteed */
	line_fun = line_score_single_category;
	postprocess_fun = NULL;
      } else {
	line_fun = NULL;
	postprocess_fun = score_single_category;
      }

    } else {
      /* multiple categories */

      preprocess_fun = NULL;
      word_fun = score_word;
      if( options & (1<<OPTION_FILTER) ) {
	empirical.track_features = 1; /* complex calc should be guaranteed */
	line_fun = line_score_multiple_categories;
	postprocess_fun = NULL;
      } else {
	line_fun = NULL;
	postprocess_fun = score_multiple_categories;
      }

    }
  } else if( options & (1<<OPTION_LEARN) ) {
    /* category learning */

    preprocess_fun = init_learner;
    word_fun = hash_word_and_learn;
    line_fun = NULL;
    postprocess_fun = optimize_learner_and_save;

  } else { /* something wrong ? */
    usage(argv);
    exit(0);
  }

  if( preprocess_fun ) { (*preprocess_fun)(); }

  /* preallocate primary text holding buffer */
  textbuf_len = BUFLEN;
  textbuf = malloc(textbuf_len);

  /* now process each file on the command line,
     or if none provided read stdin */
  while( (optind > -1) && *(argv + optind) ) {
    /* if it's a filename, process it */
    if( (input = fopen(argv[optind], "r")) ) {

      options |= (1<<INPUT_FROM_CMDLINE);

      if( (options & (1<<OPTION_VERBOSE)) && 
	  !(options & (1<<OPTION_CLASSIFY))) {
	fprintf(stdout, "processing file %s\n", argv[optind]);
      }

      /* set some initial options */
      xml.state = TEXT; 
      if( options & (1<<OPTION_MBOX_FORMAT) ) {

	mbox.state = UNDEF;
	mbox.prev_line_empty = 1;
	if( !(options & (1<<OPTION_I18N)) ) {
	  process_file(input, mbox_line_filter, 
		       ((options & (1<<OPTION_XML)) ? xml_character_filter : NULL), 
		       word_fun, line_fun);
	} else {
	  w_process_file(input, w_mbox_line_filter, 
			 ((options & (1<<OPTION_XML)) ? w_xml_character_filter : NULL), 
			 word_fun, line_fun);
	}
      } else { 
	/* default to text */
	if( !(options & (1<<OPTION_I18N)) ) {
	  process_file(input, NULL, 
		       ((options & (1<<OPTION_XML)) ? xml_character_filter : NULL), 
		       word_fun, line_fun);
	} else {
	  w_process_file(input, NULL, 
			 ((options & (1<<OPTION_XML)) ? w_xml_character_filter : NULL), 
			 word_fun, line_fun);
	}
      }

      fclose(input);

    } else { /* unrecognized file name */

      fprintf(stderr, "error: couldn't open %s\n", argv[optind]);
      usage(argv);
      exit(0);

    }
    optind++;
  }
  /* in case no files were specified, get input from stdin */
  if( !(options & (1<<INPUT_FROM_CMDLINE)) ) {

    if( (options & (1<<OPTION_VERBOSE)) && 
	!(options & (1<<OPTION_CLASSIFY)) ) {
      fprintf(stdout, "taking input from stdin\n");
    }

    /* set some initial options */
    xml.state = TEXT; 

    if( options & (1<<OPTION_MBOX_FORMAT) ) {

      mbox.state = UNDEF;
      mbox.prev_line_empty = 1;
      if( !(options & (1<<OPTION_I18N)) ) {
	process_file(stdin, mbox_line_filter, 
		     ((options & (1<<OPTION_XML)) ? xml_character_filter : NULL), 
		     word_fun, line_fun);
      } else {
	w_process_file(stdin, w_mbox_line_filter, 
		       ((options & (1<<OPTION_XML)) ? w_xml_character_filter : NULL), 
		       word_fun, line_fun);
      }

    } else { 
      /* default to text */
      if( !(options & (1<<OPTION_I18N)) ) {
	process_file(stdin, NULL,
		     ((options & (1<<OPTION_XML)) ? xml_character_filter : NULL), 
		     word_fun, line_fun);	
      } else {
	w_process_file(stdin, NULL, 
		       ((options & (1<<OPTION_XML)) ? w_xml_character_filter : NULL), 
		       word_fun, line_fun);
      }
    }

  }
  
  if( postprocess_fun ) { (*postprocess_fun)(); }

  /* free some global resources */
  free(textbuf);

  for(k = 0; k < regex_count; k++) {
    regfree(&re[k].regex);
  }

  exit(exit_code);
}
